/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Pathfinder
 * Description: -
 */

#define Uses_Area
#define Uses_Sector
#define Uses_Terrain
#define Uses_Object
#define Uses_ProgramManager
#define Uses_Util
#define Uses_Brain
#include "mheader.h"

#include "pathfind.h"

#include "path.h"



static bool			find_path(const AREA_POINT *, const AREA_POINT *);

static void			store_path(void);

static int			obstacle_at(int, int);


static const AREA_POINT *	PathStart;

static PATH_NODE *		Path = NULL;
static AREA_DISTANCE		PathLength;

static PATH_NODE *		StoredPath = NULL;
static AREA_DISTANCE		StoredPathLength;
static bool			StoredPathExists;

static const CHARACTER *	PathSubject;

static bool			IgnoreCharacters;
static bool			IgnoreDestructable;
static bool			IgnoreDangerous;



/*
 * pathfinder module init
 */
void pathfinder_init(void)
{
	Path = checked_malloc(MAX_PATH * sizeof *Path);

	StoredPath = checked_malloc(MAX_PATH * sizeof *StoredPath);

	ptf_setup(AREA_HEIGHT, AREA_WIDTH, MAX_PATH, obstacle_at);
}



/*
 * pathfinder module clean up
 */
void pathfinder_clean_up(void)
{
	ptf_clean_up();

	free(StoredPath);

	free(Path);
}



/*
 * tries to find a path (AI run version)
 */
PATH_NODE * find_run_path(const CHARACTER *character, const AREA_POINT *target)
{
#define MIN_DIFF_TO_MATTER	5
	PathSubject = character;

	StoredPathExists = false;
	IgnoreDestructable = IgnoreDangerous = false;
	IgnoreCharacters = true;

	if (find_path(&character->location, target)) {

		if (PathLength <= MAX_SHORT_PATH) return Path;

		store_path();
	}

	/* 2 */
	IgnoreDestructable = true;

	if (find_path(&character->location, target)) {

		if (!StoredPathExists) return Path;

		if (PathLength + MIN_DIFF_TO_MATTER <
			StoredPathLength) {

			return Path;
		} else {

			return StoredPath;
		}

	}

	/* 3 */
	IgnoreDangerous = true;

	if (find_path(&character->location, target)) {

		return Path;
	}

	return NULL;
}



/*
 * tries to find a path (AI run around character version)
 */
PATH_NODE * find_run_around_character_path(const CHARACTER *character,
	const AREA_POINT *target
)
{
	PathSubject = character;

	IgnoreCharacters = IgnoreDestructable =
		IgnoreDangerous = false;

	if (find_path(&character->location, target)) {

		return Path;
	}

	return NULL;
}



/*
 * tries to find a path (AI patrol version)
 */
PATH_NODE * find_patrol_path(const CHARACTER *character, const AREA_POINT *target)
{
	PathSubject = character;

	IgnoreDestructable = IgnoreDangerous = false;
	IgnoreCharacters = true;

	if (find_path(&character->location, target)) {

		return Path;
	}

	return NULL;
}



/*
 * tries to find a path (run command version)
 */
PATH_NODE * find_run_command_path(const CHARACTER *character,
	const AREA_POINT *target
)
{
	PathSubject = character;

	IgnoreCharacters = IgnoreDestructable = false;
	IgnoreDangerous = true;

	if (find_path(&character->location, target)) {

		return Path;
	}

	return NULL;
}



/*
 * tries to find a path (jump version)
 */
PATH_NODE * find_jump_path(const CHARACTER *character, const AREA_POINT *target)
{
	PathSubject = character;

	IgnoreCharacters = IgnoreDangerous = true;
	IgnoreDestructable = false;

	if (find_path(&character->location, target)) {
		return Path;
	}

	return NULL;
}



/*
 * returns the length of a path
 */
AREA_DISTANCE path_length(const PATH_NODE *node)
{
	if (node == Path) return PathLength;
	if (node == StoredPath) return StoredPathLength;

	die("*** CORE ERROR *** bug in path_length()");

	/* NEVER REACHED */
	return 0;
}



/*
 * tries to find a path
 */
static bool find_path(const AREA_POINT *start, const AREA_POINT *target)
{
	const PTF_PATH_NODE *path;
	const PTF_PATH_NODE *node;
	int i;

	PathStart = start;

	path = ptf_find_path(start->y, start->x, target->y, target->x);

	if (path == NULL) return false;

	for (node = path, i = 0;
		node != NULL && i < MAX_PATH;
		node = node->parent, i++) {

		Path[i].p.y = node->y;
		Path[i].p.x = node->x;

		if (i < MAX_PATH - 1) {

			Path[i].parent = &Path[i + 1];
		}
	}

	Path[i - 1].parent = NULL;
	PathLength = i;

	return true;
}



/*
 * stores the current path
 */
static void store_path(void)
{

	memcpy(StoredPath, Path, MAX_PATH * sizeof *Path);
	StoredPathLength = PathLength;
	StoredPathExists = true;
}



/*
 * returns 1 if there is an obstacle at the passed point
 * or 0 if there is no obstacle
 */
static int obstacle_at(int y, int x)
{
	AREA_POINT p;
	SECTOR *sector;
	TERRAIN *terrain;
	OBJECT *object;
	CHARACTER *character;

	p.y = y;
	p.x = x;
	sector = sector_at(&p);

	character = sector->character;
	if (!IgnoreCharacters && character != NULL &&
		!area_points_equal(&p, PathStart)) {

		return true;
	}

	object = sector->object;
	if (object != NULL) {

		if (object_has_attribute(object, OA_IMPASSABLE)
			&& (!IgnoreDestructable ||
			object->condition ==
			CONDITION_INDESTRUCTABLE ||
			object->condition > AI_OBSTACLE_DESTRUCTION_MAX)) {

			return true;
		}
	}

	terrain = sector->terrain;
	if (terrain != NULL) {
		const TERRAIN_DATA *terrain_data;

		terrain_data = terrain_static_data(terrain);

		if (terrain_data->attribute[TA_IMPASSABLE]) {
			return true;
		}

		if (!IgnoreDangerous &&
			terrain_data->attribute[TA_DANGEROUS]) {

			return terrain_dangerous_for(PathSubject,
				terrain
			);
		}
	}

	return false;
}

